from __future__ import annotations

import os

from prettyqt import widgets
from prettyqt.qt import QtSvg
from prettyqt.utils import types


QtSvg.QSvgWidget.__bases__ = (widgets.Widget,)


class SvgWidget(QtSvg.QSvgWidget):
    def load_file(self, path: types.PathType):
        self.load(os.fspath(path))



from __future__ import annotations

from prettyqt import core, widgets
from prettyqt.qt import QtHelp
from prettyqt.utils import types


QtHelp.QHelpSearchResultWidget.__bases__ = (widgets.Widget,)


class HelpSearchResultWidget(QtHelp.QHelpSearchResultWidget):
    def get_link_at(self, point: types.PointType) -> core.Url:
        if isinstance(point, tuple):
            point = core.Point(*point)
        return core.Url(self.linkAt(point))



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtScxml


QtScxml.QScxmlStateMachine.__bases__ = (core.Object,)


class ScxmlStateMachine(QtScxml.QScxmlStateMachine):
    pass



from __future__ import annotations

from collections.abc import Iterator
import os

from prettyqt import core
from prettyqt.qt import QtCore, QtQml
from prettyqt.utils import types


QtQml.QQmlApplicationEngine.__bases__ = (QtQml.QQmlEngine,)


class QmlApplicationEngine(QtQml.QQmlApplicationEngine):
    def __iter__(self) -> Iterator[QtCore.QObject]:
        return iter(self.rootObjects())

    def load_data(
        self,
        data: types.ByteArrayType,
        url: types.UrlType | None = None,
    ):
        if isinstance(data, str):
            data = data.encode()
        if isinstance(data, bytes):
            data = QtCore.QByteArray(data)
        if isinstance(url, str):
            url = core.Url.from_user_input(url)
        elif url is None:
            url = core.Url()
        self.loadData(data, url)

    def load_file(self, file: types.UrlType | types.PathType):
        if isinstance(file, os.PathLike):
            file = os.fspath(file)
        self.load(file)



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtPositioning
from prettyqt.utils import bidict


QtPositioning.QGeoAreaMonitorSource.__bases__ = (core.Object,)


mod = QtPositioning.QGeoAreaMonitorSource

ERRORS = bidict(
    access_error=mod.Error.AccessError,
    insufficient_pos_info=mod.Error.InsufficientPositionInfo,
    none=mod.Error.NoError,
    unknown_source=mod.Error.UnknownSourceError,
)

ErrorStr = Literal["access_error", "insufficient_pos_info", "none", "unknown_source"]

AREA_MONITOR_FEATURES = bidict(
    persistent_area=mod.AreaMonitorFeature.PersistentAreaMonitorFeature,
    any_area=mod.AreaMonitorFeature.AnyAreaMonitorFeature,
)

AreaMonitorFeatureStr = Literal["persistent_area", "any_area"]


class GeoAreaMonitorSource(QtPositioning.QGeoAreaMonitorSource):
    def __str__(self):
        return self.sourceName()

    def __repr__(self):
        return f"{type(self).__name__}({self.name()!r})"

    def get_error(self) -> AreaMonitorFeatureStr:
        """Return error type.

        Returns:
            error type
        """
        return ERRORS.inverse[self.error()]


if __name__ == "__main__":
    obj = core.Object()
    coord = GeoAreaMonitorSource(obj)
    print(str(coord))
    print(repr(coord))



"""Provides QtCore classes and functions."""

from __future__ import annotations


# import sys
from typing import TYPE_CHECKING

from prettyqt.qt import PYQT5, PYQT6, PYSIDE6

if TYPE_CHECKING:
    import datetime

# sys.setrecursionlimit(2000)


def to_datetime(self) -> datetime.datetime:
    return self.toPyDateTime()


def to_date(self) -> datetime.date:
    return self.toPyDate()


def to_time(self) -> datetime.time:
    return self.toPyTime()


if PYQT5:
    from PyQt5.QtCore import (
        QT_VERSION_STR,
        QT_VERSION_STR as __version__,
        QDateTime,
        QDate,
        QTime,
        pyqtBoundSignal,
        pyqtBoundSignal as SignalInstance,
        pyqtProperty,
        pyqtProperty as Property,
        pyqtSignal,
        pyqtSignal as Signal,
        pyqtSlot,
        pyqtSlot as Slot,
        Q_ENUM,
        Q_ENUM as QEnum,
        Q_ARG,
        Q_ARG as QGenericArgument,
        Q_FLAG,
        Q_FLAG as QFlag,
        Q_CLASSINFO,
        Q_CLASSINFO as QClassInfo,
        PYQT_VERSION_STR as BINDING_VERSION,
    )

    # For issue #153
    from PyQt5.QtCore import *

    QDateTime.toPython = to_datetime
    QDate.toPython = to_date
    QTime.toPython = to_time
    QLibraryInfo.LibraryPath = QLibraryInfo.LibraryLocation  # type: ignore
    QRegularExpression.MatchOption.AnchorAtOffsetMatchOption = (  # type: ignore
        QRegularExpression.MatchOption.AnchoredMatchOption
    )
    # Those are imported from `import *`
    del (
        pyqtSignal,
        pyqtBoundSignal,
        pyqtSlot,
        pyqtProperty,
        QT_VERSION_STR,
        Q_FLAG,
        Q_ENUM,
        Q_ARG,
        Q_CLASSINFO,
    )
elif PYQT6:
    from PyQt6.QtCore import (
        QT_VERSION_STR,
        QT_VERSION_STR as __version__,
        QDateTime,
        QDate,
        QTime,
        pyqtBoundSignal,
        pyqtBoundSignal as SignalInstance,
        pyqtProperty,
        pyqtProperty as Property,
        pyqtSignal,
        pyqtSignal as Signal,
        pyqtSlot,
        pyqtSlot as Slot,
        pyqtEnum,
        pyqtEnum as QEnum,
        pyqtEnum as QFlag,
        pyqtClassInfo,
        pyqtClassInfo as QClassInfo,
        Q_ARG,
        Q_ARG as QGenericArgument,
        PYQT_VERSION_STR as BINDING_VERSION,
        QtMsgType,
    )

    QtCriticalMsg = QtMsgType.QtCriticalMsg
    QtDebugMsg = QtMsgType.QtDebugMsg
    QtFatalMsg = QtMsgType.QtFatalMsg
    QtInfoMsg = QtMsgType.QtInfoMsg
    QtSystemMsg = QtMsgType.QtSystemMsg
    QtWarningMsg = QtMsgType.QtWarningMsg

    # For issue #153
    from PyQt6.QtCore import *

    QDateTime.toPython = to_datetime
    QDate.toPython = to_date
    QTime.toPython = to_time
    QCoreApplication.exec_ = QCoreApplication.exec  # type: ignore
    QLibraryInfo.location = QLibraryInfo.path  # type: ignore
    # Those are imported from `import *`
    del (
        pyqtSignal,
        pyqtBoundSignal,
        pyqtSlot,
        pyqtProperty,
        QT_VERSION_STR,
        pyqtEnum,
        Q_ARG,
        pyqtClassInfo,
    )
    for cls in (QEvent, Qt):
        for attr in dir(cls):
            if not attr[0].isupper():
                continue
            ns = getattr(cls, attr)
            for name, val in vars(ns).items():
                if not name.startswith("_"):
                    setattr(cls, name, val)
elif PYSIDE6:
    from PySide6.QtCore import *  # type: ignore
    from PySide6.QtCore import __version__  # type: ignore
    from PySide6 import __version__ as BINDING_VERSION



from importlib import resources
import pathlib


ROOT_PATH = pathlib.Path(resources.files("prettyqt"))  # type: ignore

LOCALIZATION_PATH = ROOT_PATH / "localization"
THEMES_PATH = ROOT_PATH / "themes"
RE_LEXER_PATH = (
    ROOT_PATH / "syntaxhighlighters" / "pygments" / "regularexpressionlexer.py"
)
ICON_FONT_PATH = ROOT_PATH / "iconprovider" / "fonts"



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui, qt
from prettyqt.qt import QtCharts
from prettyqt.utils import bidict


VALUE_POSITION = bidict(
    lower_extreme=QtCharts.QBoxSet.ValuePositions.LowerExtreme,
    lower_quartile=QtCharts.QBoxSet.ValuePositions.LowerQuartile,
    median=QtCharts.QBoxSet.ValuePositions.Median,
    upper_quartile=QtCharts.QBoxSet.ValuePositions.UpperQuartile,
    upper_extreme=QtCharts.QBoxSet.ValuePositions.UpperExtreme,
)

ValuePositionStr = Literal[
    "lower_extreme", "lower_quartile", "median", "upper_quartile", "upper_extreme"
]

QtCharts.QBoxSet.__bases__ = (core.Object,)


class BoxSet(QtCharts.QBoxSet):
    def __repr__(self):
        return (
            f"{type(self).__name__}({self['lower_extreme']}, {self['lower_quartile']},"
            f" {self['median']}, {self['upper_quartile']}, {self['upper_extreme']}, "
            f"{self.label()!r})"
        )

    def __getitem__(self, index: int | ValuePositionStr) -> float:
        val = VALUE_POSITION[index] if isinstance(index, str) else index
        if not (0 <= qt.flag_to_int(val) <= 4):
            raise KeyError(val)
        return self.at(qt.flag_to_int(val))

    def __setitem__(self, index: int | ValuePositionStr, value: int):
        val = VALUE_POSITION[index] if isinstance(index, str) else index
        if not (0 <= qt.flag_to_int(val) <= 4):
            raise KeyError(val)
        self.setValue(qt.flag_to_int(val), value)

    def get_pen(self) -> gui.Pen:
        return gui.Pen(self.pen())

    def get_brush(self) -> gui.Brush:
        return gui.Brush(self.brush())


if __name__ == "__main__":
    boxset = BoxSet()
    print(repr(boxset))



from __future__ import annotations

import mimetypes

from prettyqt import gui, iconprovider, widgets
from prettyqt.qt import QtCore, QtWidgets


class AwesomeFileIconProvider(widgets.FileIconProvider):
    def __init__(self):
        super().__init__()
        self.folder_icon = iconprovider.get_icon("mdi.folder")
        self.text_icon = iconprovider.get_icon("mdi.file-document")
        self.file_icon = iconprovider.get_icon("mdi.file")
        self.desktop_icon = iconprovider.get_icon("mdi.desktop-mac-dashboard")
        self.computer_icon = iconprovider.get_icon("mdi.desktop-classic")
        self.trashcan_icon = iconprovider.get_icon("mdi.trash-can")
        self.drive_icon = iconprovider.get_icon("mdi.harddisk")
        self.network_icon = iconprovider.get_icon("mdi.folder-network")

    def icon(
        self, icon: QtWidgets.QFileIconProvider.IconType | QtCore.QFileInfo
    ) -> gui.Icon:
        if isinstance(icon, QtCore.QFileInfo):
            if icon.isDir():
                return self.folder_icon
            elif icon.isFile():
                return self.file_icon
        elif icon == widgets.FileIconProvider.IconType.Folder:
            return self.folder_icon
        # elif icon == widgets.FileIconProvider.IconType.File:
        #     return self.file_icon
        elif icon == widgets.FileIconProvider.IconType.Computer:
            return self.computer_icon
        elif icon == widgets.FileIconProvider.IconType.Desktop:
            return self.desktop_icon
        elif icon == widgets.FileIconProvider.IconType.Trashcan:
            return self.trashcan_icon
        elif icon == widgets.FileIconProvider.IconType.Network:
            return self.network_icon
        elif icon == widgets.FileIconProvider.IconType.Drive:
            return self.drive_icon
        return self.file_icon

    @staticmethod
    def mimetype_icon(path, fallback=None):
        """Try to create an icon from theme using the file mimetype.

        E.g.::

            return self.mimetype_icon(
                path, fallback=':/icons/text-x-python.png')

        :param path: file path for which the icon must be created
        :param fallback: fallback icon path (qrc or file system)
        :returns: QIcon or None if the file mimetype icon could not be found.
        """
        mime = mimetypes.guess_type(path)[0]
        if mime:
            icon = mime.replace("/", "-")
            # if system.WINDOWS:
            #     return icons.file()
            if gui.Icon.hasThemeIcon(icon):
                icon = gui.Icon.fromTheme(icon)
                if not icon.isNull():
                    return icon
        if fallback:
            return gui.Icon(fallback)
        return gui.Icon.fromTheme("text-x-generic")


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    app.load_language("de")
    model = widgets.FileSystemModel()
    model.set_root_path("root")
    prov = AwesomeFileIconProvider()
    model.setIconProvider(prov)
    tree = widgets.TreeView()
    tree.set_model(model)
    tree.show()
    app.main_loop()



from __future__ import annotations

import os
from typing import Literal

from prettyqt import constants, gui, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict, colors, helpers, types


TEXT_INTERACTION = bidict(
    none=QtCore.Qt.TextInteractionFlag.NoTextInteraction,
    by_mouse=QtCore.Qt.TextInteractionFlag.TextSelectableByMouse,
    by_keyboard=QtCore.Qt.TextInteractionFlag.TextSelectableByKeyboard,
    accessible_by_mouse=QtCore.Qt.TextInteractionFlag.LinksAccessibleByMouse,
    accessible_by_keyboard=QtCore.Qt.TextInteractionFlag.LinksAccessibleByKeyboard,
    text_editable=QtCore.Qt.TextInteractionFlag.TextEditable,
    like_text_editor=QtCore.Qt.TextInteractionFlag.TextEditorInteraction,
    like_text_browser=QtCore.Qt.TextInteractionFlag.TextBrowserInteraction,
)

TextInteractionStr = Literal[
    "none",
    "by_mouse",
    "by_keyboard",
    "accessible_by_mouse",
    "accessible_by_keyboard",
    "text_editable",
    "like_text_editor",
    "like_text_browser",
]

TEXT_FORMAT = bidict(
    rich=QtCore.Qt.TextFormat.RichText,
    plain=QtCore.Qt.TextFormat.PlainText,
    auto=QtCore.Qt.TextFormat.AutoText,
    markdown=QtCore.Qt.TextFormat.MarkdownText,
)


TextFormatStr = Literal["rich", "plain", "auto", "markdown"]


QtWidgets.QLabel.__bases__ = (widgets.Frame,)


class Label(QtWidgets.QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.openExternalLinks()

    def __repr__(self):
        return f"{type(self).__name__}({self.text()!r})"

    def serialize_fields(self):
        pixmap = gui.Pixmap(self.pixmap())
        return dict(
            text=self.text(),
            scaled_contents=self.hasScaledContents(),
            indent=self.indent(),
            margin=self.margin(),
            text_format=self.get_text_format(),
            pixmap=pixmap if pixmap else None,
            open_external_links=self.openExternalLinks(),
            has_selected_text=self.hasSelectedText(),
            selected_text=self.selectedText(),
            alignment=(self.get_horizontal_alignment(), self.get_vertical_alignment()),
            word_wrap=self.wordWrap(),
            text_interaction_flags=self.get_text_interaction(),
        )

    def __setstate__(self, state):
        self.setText(state.get("text", ""))
        self.setIndent(state.get("indent", -1))
        self.setMargin(state.get("margin", 0))
        self.setWordWrap(state.get("word_wrap", 0))
        self.set_text_format(state.get("text_format", 0))
        # self.setPixmap(state.get("pixmap"))
        self.setOpenExternalLinks(state.get("open_external_links", False))
        self.set_alignment(*state.get("alignment"))
        self.setScaledContents(state["scaled_contents"])
        self.setWordWrap(state["word_wrap"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def allow_links(self) -> Label:
        # self.setText("<a href=\"http://example.com/\">Click Here!</a>")
        self.setTextFormat(QtCore.Qt.TextFormat.RichText)
        self.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextBrowserInteraction  # type: ignore
        )
        self.setOpenExternalLinks(True)
        return self

    def set_alignment(
        self,
        horizontal: constants.HorizontalAlignmentStr | None = None,
        vertical: constants.VerticalAlignmentStr | None = None,
    ):
        """Set the alignment of the label's contents."""
        if horizontal is None and vertical is not None:
            flag = constants.V_ALIGNMENT[vertical]
        elif vertical is None and horizontal is not None:
            flag = constants.H_ALIGNMENT[horizontal]
        elif vertical is not None and horizontal is not None:
            flag = constants.V_ALIGNMENT[vertical] | constants.H_ALIGNMENT[horizontal]
        else:
            return
        self.setAlignment(flag)
        return self

    def get_horizontal_alignment(self) -> constants.HorizontalAlignmentStr:
        align = self.alignment()
        if align & constants.ALIGN_RIGHT:  # type: ignore
            return "right"
        elif align & constants.ALIGN_H_CENTER:  # type: ignore
            return "center"
        elif align & constants.ALIGN_JUSTIFY:  # type: ignore
            return "justify"
        else:
            return "left"

    def get_vertical_alignment(self) -> constants.VerticalAlignmentStr:
        align = self.alignment()
        if align & constants.ALIGN_TOP:  # type: ignore
            return "top"
        elif align & constants.ALIGN_BOTTOM:  # type: ignore
            return "bottom"
        elif align & constants.ALIGN_BASELINE:  # type: ignore
            return "baseline"
        else:
            return "center"

    def set_indent(self, indent: int) -> Label:
        """Set the label's text indent in pixels."""
        self.setIndent(indent)
        return self

    def set_text_format(self, text_format: TextFormatStr) -> Label:
        """Set the text format.

        Args:
            text_format: text format to use

        Raises:
            InvalidParamError: text format does not exist
        """
        if text_format not in TEXT_FORMAT:
            raise InvalidParamError(text_format, TEXT_FORMAT)
        self.setTextFormat(TEXT_FORMAT[text_format])
        return self

    def get_text_format(self) -> TextFormatStr:
        """Return current text format.

        Returns:
            text format
        """
        return TEXT_FORMAT.inverse[self.textFormat()]

    def set_text_interaction(self, *types: TextInteractionStr) -> Label:
        """Set the text interaction mode.

        Args:
            types: text interaction mode to use

        Raises:
            InvalidParamError: text interaction mode does not exist
        """
        for item in types:
            if item not in TEXT_INTERACTION:
                raise InvalidParamError(item, TEXT_INTERACTION)
        flags = helpers.merge_flags(types, TEXT_INTERACTION)
        self.setTextInteractionFlags(flags)
        return self

    def get_text_interaction(self) -> list[TextInteractionStr]:
        """Return current text interaction mode.

        Returns:
            list of text interaction modes
        """
        return [k for k, v in TEXT_INTERACTION.items() if v & self.textInteractionFlags()]

    def set_text(self, text: str) -> Label:
        """Set the label's text."""
        self.setText(text)
        return self

    def set_bold(self, bold: bool = True) -> Label:
        font = self.font()
        font.setBold(bold)
        self.setFont(font)
        return self

    def set_italic(self, italic: bool = True) -> Label:
        font = self.font()
        font.setItalic(italic)
        self.setFont(font)
        return self

    def set_point_size(self, size: int) -> Label:
        font = self.font()
        font.setPointSize(size)
        self.setFont(font)
        return self

    def set_weight(self, weight: gui.font.WeightStr) -> Label:
        """Set the font weight.

        Args:
            weight: font weight

        Raises:
            InvalidParamError: invalid font weight
        """
        if weight not in gui.font.WEIGHT:
            raise InvalidParamError(weight, gui.font.WEIGHT)
        font = self.font()
        font.setWeight(gui.font.WEIGHT[weight])
        self.setFont(font)
        return self

    def set_color(self, color: types.ColorType) -> Label:
        with self.edit_stylesheet() as ss:
            if color is None:
                ss.color.setValue("")
            else:
                color = colors.get_color(color)
                ss.color.setValue(color.name())
        return self

    def set_image(self, path: types.PathType, width: int = 300) -> Label:
        self.setScaledContents(True)
        self.set_alignment(horizontal="center")
        self.setText(
            "<html><head/><body><p>"
            f"<img src={os.fspath(path)!r} width={str(width)!r}/>"
            "</p></body></html>"
        )
        return self

    @classmethod
    def image_from_path(
        cls, path: types.PathType, parent: QtWidgets.QWidget | None = None
    ) -> Label:
        pixmap = gui.Pixmap.from_file(path)
        label = cls(parent=parent)
        label.setPixmap(pixmap)
        label.resize(pixmap.width(), pixmap.height())
        return label


if __name__ == "__main__":
    app = widgets.app()
    widget = Label("http://www.test.de")
    widget.show()
    app.main_loop()



from __future__ import annotations

from typing import Callable

from prettyqt import constants, core
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError, helpers


QtCore.QTimer.__bases__ = (core.Object,)


class Timer(QtCore.QTimer):
    def serialize_fields(self):
        return dict(
            interval=self.interval(),
            single_shot=self.isSingleShot(),
            timer_type=self.get_type(),
        )

    @classmethod
    def single_shot(cls, callback: Callable) -> Timer:
        timer = cls()
        timer.timeout.connect(callback)
        timer.setSingleShot(True)
        return timer

    def set_type(self, typ: constants.TimerTypeStr):
        """Set the timer type.

        Args:
            typ: timer type

        Raises:
            InvalidParamError: timer type does not exist
        """
        if typ not in constants.TIMER_TYPE:
            raise InvalidParamError(typ, constants.TIMER_TYPE)
        self.setTimerType(constants.TIMER_TYPE[typ])

    def get_type(self) -> constants.TimerTypeStr:
        """Return current timer type.

        Returns:
            timer type
        """
        return constants.TIMER_TYPE.inverse[self.timerType()]

    def set_interval(self, interval: int | str):
        if isinstance(interval, str):
            interval = helpers.parse_time(interval)
        self.setInterval(interval)

    def start_timer(self, interval: None | int | str = None):
        if isinstance(interval, str):
            interval = helpers.parse_time(interval)
        if interval is None:
            self.start()
        else:
            self.start(interval)

    def restart(self):
        self.stop()
        self.start()


if __name__ == "__main__":
    timer = Timer()
    timer.set_interval("2m")



"""Provides QtStateMachine classes and functions."""

from prettyqt.qt import PYQT5, PYQT6, PythonQtError


if PYQT5:
    from PyQt5.Qsci import *
elif PYQT6:
    from PyQt6.Qsci import *
else:
    raise PythonQtError("No Qt bindings could be found")



from __future__ import annotations

from prettyqt import core, gui, widgets
from prettyqt.qt import QtWidgets


class ElidedLabel(widgets.Frame):
    elision_changed = core.Signal(bool)

    def __init__(
        self,
        text: str = "",
        parent: QtWidgets.QWidget | None = None,
    ):
        super().__init__(parent=parent)
        self.elided = False
        self.content = text
        self.set_size_policy("expanding", "preferred")

    def __repr__(self):
        return f"{type(self).__name__}({self.text()!r})"

    def set_text(self, text: str):
        self.content = text
        self.update()

    def paintEvent(self, event):
        super().paintEvent(event)
        painter = gui.Painter(self)
        metrics = painter.get_font_metrics()
        did_elide = False
        line_spacing = metrics.lineSpacing()
        y = 0
        layout = gui.TextLayout(self.content, painter.font())
        with layout.process_layout():
            while True:
                line = layout.createLine()

                if not line.isValid():
                    break

                line.setLineWidth(self.width())
                next_line_y = y + line_spacing

                if self.height() >= next_line_y + line_spacing:
                    line.draw(painter, core.Point(0, y))
                    y = next_line_y
                else:
                    last_line = self.content[line.textStart() :]
                    elided_line = metrics.elided_text(last_line, "right", self.width())
                    painter.drawText(0, y + metrics.ascent(), elided_line)
                    line = layout.createLine()
                    did_elide = line.isValid()
                    break
        if did_elide != self.elided:
            self.elided = did_elide
            self.elision_changed.emit(did_elide)

    # def paintEvent(self, event):
    #     painter = gui.Painter(self)
    #     metrics = gui.FontMetrics(self.font())
    #     elided = metrics.elided_text(self.text(), "right", self.width())
    #     painter.drawText(self.rect(), self.alignment(), elided)


if __name__ == "__main__":
    app = widgets.app()
    widget = ElidedLabel("test")
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import statemachine
from prettyqt.qt import QtStateMachine


QtStateMachine.QSignalTransition.__bases__ = (statemachine.AbstractTransition,)


class SignalTransition(QtStateMachine.QSignalTransition):
    pass



from __future__ import annotations

from prettyqt import statemachine
from prettyqt.qt import QtStateMachine


QtStateMachine.QMouseEventTransition.__bases__ = (statemachine.EventTransition,)


class MouseEventTransition(QtStateMachine.QMouseEventTransition):
    pass



from __future__ import annotations

import os
from typing import Literal

from prettyqt import core, gui
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import bidict, types


mod = QtWidgets.QFileIconProvider

ICON_TYPE = bidict(
    computer=mod.IconType.Computer,
    desktop=mod.IconType.Desktop,
    trashcan=mod.IconType.Trashcan,
    network=mod.IconType.Network,
    drive=mod.IconType.Drive,
    folder=mod.IconType.Folder,
    file=mod.IconType.File,
)

IconTypeStr = Literal[
    "computer", "desktop", "trashcan", "network", "drive", "folder", "file"
]

OPTIONS = bidict(
    dont_use_custom_dir_icons=mod.Option.DontUseCustomDirectoryIcons,
)

OptionStr = Literal["dont_use_custom_dir_icons"]


class FileIconProvider(QtWidgets.QFileIconProvider):
    def get_icon(self, typ: IconTypeStr | QtCore.QFileInfo | types.PathType) -> gui.Icon:
        if isinstance(typ, (os.PathLike, QtCore.QFileInfo)):
            param = core.FileInfo(typ)
        else:
            param = ICON_TYPE[typ]
        return gui.Icon(self.icon(param))


if __name__ == "__main__":
    import pathlib

    from prettyqt import widgets

    app = widgets.app()
    path = pathlib.Path.home()
    provider = FileIconProvider()
    provider.get_icon(path)



from __future__ import annotations

from dataclasses import dataclass
import enum
from typing import Literal

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtWidgets


@dataclass
class ItemWrapper:
    item: QtWidgets.QWidgetItem
    position: BorderLayout.Position


class BorderLayout(widgets.Layout):
    class Position(enum.IntEnum):
        """Item position."""

        West = 0
        North = 1
        South = 2
        East = 3
        Center = 4

    QtCore.QEnum(Position)

    def __init__(
        self,
        parent: QtWidgets.QWidget | None = None,
        margin: int = 0,
        spacing: int | None = None,
    ):
        super().__init__(parent)  # type: ignore
        self.set_margin(margin)
        self.setSpacing(spacing if spacing is not None else -1)
        self.items: list[ItemWrapper] = []

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item: QtWidgets.QWidgetItem):
        self.add_widgetitem(item, BorderLayout.Position.West)

    def addWidget(
        self,
        widget: QtWidgets.QWidget,
        position: Position = Position.West,
    ):
        self.add_widgetitem(widgets.WidgetItem(widget), position)

    def expandingDirections(self):
        return constants.HORIZONTAL | constants.VERTICAL  # type: ignore

    def hasHeightForWidth(self) -> bool:
        return False

    def count(self) -> int:
        return len(self.items)

    def itemAt(self, index: int) -> QtWidgets.QWidgetItem | None:  # type: ignore
        if index < len(self.items):
            return self.items[index].item
        return None

    def minimumSize(self):
        return self.calculate_size("minimum")

    def setGeometry(self, rect: QtCore.QRect):
        center = None
        east_width = 0
        west_width = 0
        north_height = 0
        south_height = 0

        super().setGeometry(rect)

        for wrapper in self.items:
            item = wrapper.item
            position = wrapper.position

            if position == BorderLayout.Position.North:
                geom = core.Rect(
                    rect.x(), north_height, rect.width(), item.sizeHint().height()
                )
                item.setGeometry(geom)

                north_height += item.geometry().height() + self.spacing()

            elif position == BorderLayout.Position.South:
                geom = core.Rect(
                    item.geometry().x(),
                    item.geometry().y(),
                    rect.width(),
                    item.sizeHint().height(),
                )
                item.setGeometry(geom)

                south_height += item.geometry().height() + self.spacing()
                geom = core.Rect(
                    rect.x(),
                    rect.y() + rect.height() - south_height + self.spacing(),
                    item.geometry().width(),
                    item.geometry().height(),
                )
                item.setGeometry(geom)

            elif position == BorderLayout.Position.Center:
                center = wrapper

        center_height = rect.height() - north_height - south_height

        for wrapper in self.items:
            item = wrapper.item
            position = wrapper.position

            if position == BorderLayout.Position.West:
                geom = core.Rect(
                    rect.x() + west_width,
                    north_height,
                    item.sizeHint().width(),
                    center_height,
                )
                item.setGeometry(geom)

                west_width += item.geometry().width() + self.spacing()

            elif position == BorderLayout.Position.East:
                geom = core.Rect(
                    item.geometry().x(),
                    item.geometry().y(),
                    item.sizeHint().width(),
                    center_height,
                )
                item.setGeometry(geom)

                east_width += item.geometry().width() + self.spacing()

                geom = core.Rect(
                    rect.x() + rect.width() - east_width + self.spacing(),
                    north_height,
                    item.geometry().width(),
                    item.geometry().height(),
                )
                item.setGeometry(geom)

        if center:
            rect = core.Rect(
                west_width,
                north_height,
                rect.width() - east_width - west_width,
                center_height,
            )
            center.item.setGeometry(rect)

    def sizeHint(self) -> core.Size:
        return self.calculate_size("size_hint")

    def takeAt(self, index: int) -> QtWidgets.QWidgetItem | None:  # type: ignore
        if 0 <= index < len(self.items):
            layout_struct = self.items.pop(index)
            return layout_struct.item

        return None

    def add_widgetitem(self, item: QtWidgets.QWidgetItem, position: Position):
        self.items.append(ItemWrapper(item, position))

    def calculate_size(self, size_type: Literal["minimum", "size_hint"]) -> core.Size:
        total_size = core.Size()

        for wrapper in self.items:
            position = wrapper.position
            if size_type == "minimum":
                item_size = wrapper.item.minimumSize()
            else:  # size_type == "size_hint"
                item_size = wrapper.item.sizeHint()

            if position in (
                BorderLayout.Position.North,
                BorderLayout.Position.South,
                BorderLayout.Position.Center,
            ):
                total_size.setHeight(total_size.height() + item_size.height())

            if position in (
                BorderLayout.Position.West,
                BorderLayout.Position.East,
                BorderLayout.Position.Center,
            ):
                total_size.setWidth(total_size.width() + item_size.width())

        return total_size


if __name__ == "__main__":

    class Window(widgets.Widget):
        def __init__(self):
            super().__init__()

            central_widget = widgets.TextBrowser()
            central_widget.setPlainText("Central widget")

            layout = BorderLayout()
            layout.addWidget(central_widget, BorderLayout.Position.Center)

            # Because BorderLayout doesn't call its super-class addWidget() it
            # doesn't take ownership of the widgets until setLayout() is called.
            # Therefore we keep a local reference to each label to prevent it being
            # garbage collected too soon.
            label_n = self.create_label("North")
            layout.addWidget(label_n, BorderLayout.Position.North)

            label_w = self.create_label("West")
            layout.addWidget(label_w, BorderLayout.Position.West)

            label_e1 = self.create_label("East 1")
            layout.addWidget(label_e1, BorderLayout.Position.East)

            label_e2 = self.create_label("East 2")
            layout.addWidget(label_e2, BorderLayout.Position.East)

            label_s = self.create_label("South")
            layout.addWidget(label_s, BorderLayout.Position.South)

            self.setLayout(layout)

            self.setWindowTitle("Border Layout")

        @staticmethod
        def create_label(text: str):
            label = widgets.Label(text)
            label.setFrameStyle(widgets.Frame.Box | widgets.Frame.Raised)
            return label

    app = widgets.app()
    window = Window()
    window.show()
    app.main_loop()



"""Provides QtWebEngineCore classes and functions."""

from prettyqt.qt import PYQT5, PYQT6, PYSIDE6, PythonQtError


if PYQT5:
    from PyQt5.QtWebEngineCore import *
    from PyQt5.QtWebEngineWidgets import (
        QWebEngineHistory,
        QWebEngineHistoryItem,
        QWebEngineScript,
        QWebEngineScriptCollection,
        QWebEngineProfile,
        QWebEngineSettings,
        QWebEnginePage,
        QWebEngineDownloadItem as QWebEngineDownloadRequest,
        QWebEngineContextMenuData as QWebEngineContextMenuRequest,
    )
elif PYQT6:
    from PyQt6.QtWebEngineCore import *
elif PYSIDE6:
    from PySide6.QtWebEngineCore import *
else:
    raise PythonQtError("No Qt bindings could be found")



from __future__ import annotations

from prettyqt import gui
from prettyqt.qt import QtCore


class IntegerValidator(gui.Validator):
    def __init__(self, parent: QtCore.QObject | None = None):
        super().__init__(parent)
        self.range: tuple[int | None, int | None] = (None, None)

    def set_range(self, lower: int | None, upper: int | None):
        self.range = (lower, upper)

    def validate(self, text: str, pos: int):
        if not text.lstrip("-"):
            return self.State.Intermediate, text, len(text)
        if text.lstrip("-").isnumeric():
            left = float("-inf") if self.range[0] is None else self.range[0]
            right = float("inf") if self.range[1] is None else self.range[1]
            if left <= int(text) <= right:
                return self.State.Acceptable, text, len(text)
        return self.State.Invalid, text, len(text)


if __name__ == "__main__":
    from prettyqt import widgets

    val = IntegerValidator()
    app = widgets.app()
    widget = widgets.LineEdit("This is a test")
    widget.setValidator(val)
    widget.show()
    app.main_loop()



"""Provides QtStateMachine classes and functions."""

from prettyqt.qt import PYQT5, PYQT6, PYSIDE6, PythonQtError


if PYQT5:
    from PyQt5.QtCore import (
        QAbstractState,
        QAbstractTransition,
        QEventTransition,
        QFinalState,
        QHistoryState,
        QSignalTransition,
        QState,
        QStateMachine,
    )
    from PyQt5.QtWidgets import QKeyEventTransition, QMouseEventTransition
# elif PYQT6:
#     from PyQt6.QtStateMachine import *
elif PYSIDE6:
    from PySide6.QtStateMachine import *
else:
    raise PythonQtError("No Qt bindings could be found")



"""Provides Qt init stuff."""

from typing import Callable, Literal
import operator
import os
import importlib


class PythonQtError(ImportError):
    pass


def set_env_vars(qt_binding: Literal["PyQt5", "PyQt6", "PySide6"]):
    ENV_VARS = ["QT_API", "USE_QT_API", "PYTEST_QT_API", "PYQTGRAPH_QT_LIB"]
    for var in ENV_VARS:
        os.environ[var] = qt_binding


# the order in the tuple represents the search order when importing the package
packages = {
    "pyside6": "PySide6",
    "pyqt6": "PyQt6",
    "pyqt5": "PyQt5",
}

qt_api = os.getenv("QT_API")

# one can create a QT_API environment variable to select which Qt API to use
if qt_api:
    qt_api = qt_api.lower()
    if qt_api not in packages.keys():
        raise ValueError(f"Invalid QT_API environment variable {qt_api!r}")
else:
    for name, pkg_name in packages.items():
        try:
            importlib.import_module(pkg_name)
        except ImportError:
            continue
        else:
            qt_api = name
            break

if not qt_api:
    pkg_str = ", ".join(packages)
    raise ImportError(f"One of the following Qt packages must be installed: {pkg_str}")

# os.environ.setdefault("QT_API", "pyqt6")

API = qt_api
QT_VERSION = 6 if API.endswith("6") else 5

PYQT5 = API == "pyqt5"
PYQT6 = API == "pyqt6"
PYSIDE6 = API == "pyside6"


flag_to_int: Callable = (
    operator.attrgetter("value") if API.endswith("6") else int  # type: ignore
)



"""Provides QtScxml classes and functions."""

from prettyqt.qt import PYQT5, PYQT6, PYSIDE6, PythonQtError


if PYQT5:
    from PyQt5.QtScxml import *
elif PYQT6:
    from PyQt6.QtScxml import *
elif PYSIDE6:
    from PySide6.QtScxml import *
else:
    raise PythonQtError("No Qt bindings could be found")



# parts taken from https://github.com/mustafaahci/FramelessWindow

from __future__ import annotations

import ctypes
from ctypes.wintypes import LONG
import sys
from typing import Literal


try:
    import win32api
    from win32con import (
        GWL_STYLE,
        HTBOTTOM,
        HTBOTTOMLEFT,
        HTBOTTOMRIGHT,
        HTCAPTION,
        HTLEFT,
        HTRIGHT,
        HTTOP,
        HTTOPLEFT,
        HTTOPRIGHT,
        WM_NCCALCSIZE,
        WM_NCHITTEST,
        WS_CAPTION,
        WS_MAXIMIZEBOX,
        WS_MINIMIZEBOX,
        WS_POPUP,
        WS_SYSMENU,
        WS_THICKFRAME,
    )
    import win32gui
except ImportError:
    pass

from prettyqt import widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets


SideStr = Literal["left", "top", "right", "bottom"]


class TitleBarIcon(widgets.PushButton):
    def __init__(self, text, parent=None):
        super().__init__(text, parent)

        self.setStyleSheet(
            "margin: 0;" "padding: 0px;" "font-size: 16px;" "width: 44px;" "height: 30px;"
        )

        self.set_margin(0)


class CustomTitleBar(widgets.Frame):
    def __init__(self, window_widget: QtWidgets.QWidget):
        super().__init__(window_widget)

        self.window_widget = window_widget
        self.minimize_button = TitleBarIcon("🗕")
        self.maximize_button = TitleBarIcon("🗖")
        self.exit_button = TitleBarIcon("✕")
        self.setObjectName("ControlWidget")
        state = QtCore.Qt.WindowState.WindowMinimized
        self.minimize_button.clicked.connect(
            lambda: window_widget.setWindowState(state)  # type: ignore
        )
        self.maximize_button.clicked.connect(
            lambda: (
                window_widget.showNormal()
                if window_widget.isMaximized()
                else window_widget.showMaximized()
            )
        )
        self.exit_button.clicked.connect(window_widget.close)

        self.set_layout("horizontal")
        spacer_item = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding)
        self.box.addSpacerItem(spacer_item)
        for widget in [
            self.minimize_button,
            self.maximize_button,
            self.exit_button,
        ]:
            self.box.addWidget(widget)

        self.setStyleSheet("width: 100%;" "padding: 0;" "margin: 0;")
        self.setContentsMargins(0, 0, 0, 0)
        self.box.setContentsMargins(0, 0, 0, 0)
        self.box.setSpacing(0)

    def mousePressEvent(self, a0: QtGui.QMouseEvent) -> None:
        self.window_widget.windowHandle().startSystemMove()


class FramelessWindow(widgets.Widget):
    BORDER_WIDTH = 5

    def __init__(self, parent: QtWidgets.QWidget | None = None):
        super().__init__(parent)

        # Remove window title bar and frame
        self.setWindowFlags(
            QtCore.Qt.WindowFlag.Window  # type: ignore
            | QtCore.Qt.WindowFlag.FramelessWindowHint
        )

        self.title_bar = CustomTitleBar(self)
        self.main_widget = widgets.MainWindow()

        # Set up layout
        self.main_layout = widgets.BoxLayout("vertical")
        self.main_layout.addWidget(self.title_bar)
        self.main_layout.addWidget(self.main_widget)

        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)

        self.grip_layout = widgets.GridLayout()

        self.grip_layout.addLayout(self.main_layout, 1, 1)
        self.grip_layout.addWidget(EdgeGrip("top"), 0, 1)
        self.grip_layout.addWidget(EdgeGrip("right"), 1, 2)
        self.grip_layout.addWidget(EdgeGrip("bottom"), 2, 1)
        self.grip_layout.addWidget(EdgeGrip("left"), 1, 0)
        self.grip_layout.addWidget(EdgeGrip("top_left"), 0, 0)
        self.grip_layout.addWidget(EdgeGrip("top_right"), 0, 2)
        self.grip_layout.addWidget(EdgeGrip("bottom_left"), 2, 0)
        self.grip_layout.addWidget(EdgeGrip("bottom_right"), 2, 2)
        self.grip_layout.setContentsMargins(0, 0, 0, 0)
        self.grip_layout.setSpacing(0)
        self.setLayout(self.grip_layout)

        if sys.platform == "win32":

            self.hwnd = self.winId().__int__()
            window_style = win32gui.GetWindowLong(self.hwnd, GWL_STYLE)
            win32gui.SetWindowLong(
                self.hwnd,
                GWL_STYLE,
                window_style
                | WS_POPUP
                | WS_THICKFRAME
                | WS_CAPTION
                | WS_SYSMENU
                | WS_MAXIMIZEBOX
                | WS_MINIMIZEBOX,
            )

            # if QtWin.isCompositionEnabled():
            #     # Aero Shadow
            #     QtWin.extendFrameIntoClientArea(self, -1, -1, -1, -1)
            # else:
            #     QtWin.resetExtendedFrame(self)

    def __getattr__(self, attr: str):
        return getattr(self.main_widget, attr)

    def changeEvent(self, event):
        # not sure if this should be done on non-windows
        if event.type() == event.WindowStateChange:
            if self.windowState() & QtCore.Qt.WindowState.WindowMaximized:  # type: ignore
                margin = abs(self.mapToGlobal(self.rect().topLeft()).y())
                self.setContentsMargins(margin, margin, margin, margin)
            else:
                self.setContentsMargins(0, 0, 0, 0)

        return super().changeEvent(event)

    def nativeEvent(self, event, message):
        return_value, result = super().nativeEvent(event, message)
        if sys.platform != "win32":
            return return_value, result
        # if you use Windows OS
        if event == b"windows_generic_MSG":
            msg = ctypes.wintypes.MSG.from_address(message.__int__())
            # Get the coordinates when the mouse moves.
            x = win32api.LOWORD(LONG(msg.lParam).value)  # type: ignore
            # converted an unsigned int to int (for dual monitor issue)
            if x & 32768:
                x = x | -65536
            y = win32api.HIWORD(LONG(msg.lParam).value)  # type: ignore
            if y & 32768:
                y = y | -65536

            x -= self.frameGeometry().x()
            y -= self.frameGeometry().y()

            # Determine whether there are other widgets at the mouse position.
            if self.childAt(x, y) is not None and self.childAt(
                x, y
            ) is not self.findChild(widgets.Widget, "ControlWidget"):
                # passing
                if (
                    self.width() - self.BORDER_WIDTH > x > self.BORDER_WIDTH
                    and y < self.height() - self.BORDER_WIDTH
                ):
                    return return_value, result

            if msg.message == WM_NCCALCSIZE:
                # Remove system title
                return True, 0

            if msg.message == WM_NCHITTEST:
                w, h = self.width(), self.height()
                lx = x < self.BORDER_WIDTH
                rx = x > w - self.BORDER_WIDTH
                ty = y < self.BORDER_WIDTH
                by = y > h - self.BORDER_WIDTH
                if lx and ty:
                    return True, HTTOPLEFT
                if rx and by:
                    return True, HTBOTTOMRIGHT
                if rx and ty:
                    return True, HTTOPRIGHT
                if lx and by:
                    return True, HTBOTTOMLEFT
                if ty:
                    return True, HTTOP
                if by:
                    return True, HTBOTTOM
                if lx:
                    return True, HTLEFT
                if rx:
                    return True, HTRIGHT
                # Title
                return True, HTCAPTION

        return return_value, result


class EdgeGrip(widgets.Widget):
    def __init__(self, edges: str, grip_size=6, parent=None):
        super().__init__(parent)
        self.edges = edges
        self.grip_size = grip_size
        # Sides
        if edges == "top":
            self.setCursor(QtCore.Qt.SizeVerCursor)
            self.setFixedHeight(self.grip_size)
        elif edges == "right":
            self.setCursor(QtCore.Qt.SizeHorCursor)
            self.setFixedWidth(self.grip_size)
        elif edges == "bottom":
            self.setCursor(QtCore.Qt.SizeVerCursor)
            self.setFixedHeight(self.grip_size)
        elif edges == "left":
            self.setCursor(QtCore.Qt.SizeHorCursor)
            self.setFixedWidth(self.grip_size)
        # Corners
        elif edges == "top_left":
            self.setCursor(QtCore.Qt.SizeFDiagCursor)
        elif edges == "top_right":
            self.setCursor(QtCore.Qt.SizeBDiagCursor)
        elif edges == "bottom_left":
            self.setCursor(QtCore.Qt.SizeBDiagCursor)
        elif edges == "bottom_right":
            self.setCursor(QtCore.Qt.SizeFDiagCursor)

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.parent().windowHandle().startSystemResize(self.edges)


if __name__ == "__main__":
    app = widgets.app()
    m = FramelessWindow()
    button = widgets.PushButton("test")
    m.set_widget(button)
    m.show()
    m.resize(240, 160)
    app.main_loop()



from __future__ import annotations

from prettyqt import statemachine
from prettyqt.qt import QtStateMachine


QtStateMachine.QFinalState.__bases__ = (statemachine.AbstractState,)


class FinalState(QtStateMachine.QFinalState):
    pass



from __future__ import annotations

from prettyqt import gui
from prettyqt.qt import QtGui


QtGui.QRasterWindow.__bases__ = (gui.PaintDeviceWindow,)


class RasterWindow(QtGui.QRasterWindow):
    pass


if __name__ == "__main__":
    app = gui.GuiApplication([])



"""Provides QtQml classes and functions."""

from prettyqt.qt import PYQT5, PYQT6, PYSIDE6, PythonQtError


if PYQT5:
    from PyQt5.QtQml import *
elif PYQT6:
    from PyQt6.QtQml import *
elif PYSIDE6:
    from PySide6.QtQml import *  # type: ignore
else:
    raise PythonQtError("No Qt bindings could be found")



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError, bidict, types


MODE = bidict(
    normal=QtGui.QIcon.Mode.Normal,
    disabled=QtGui.QIcon.Mode.Disabled,
    active=QtGui.QIcon.Mode.Active,
    selected=QtGui.QIcon.Mode.Selected,
)

ModeStr = Literal["normal", "disabled", "active", "selected"]

STATE = bidict(off=QtGui.QIcon.State.Off, on=QtGui.QIcon.State.On)

StateStr = Literal["off", "on"]


class Icon(QtGui.QIcon):
    def __repr__(self):
        return f"{type(self).__name__}()"

    def __bool__(self):
        return not self.isNull()

    def __getstate__(self):
        pixmap = self.pixmap(256, 256)
        return bytes(core.DataStream.create_bytearray(pixmap))

    def __setstate__(self, ba):
        px = QtGui.QPixmap()
        core.DataStream.write_bytearray(ba, px)
        super().__init__(px)

    @classmethod
    def for_color(cls, color_str: str) -> Icon:
        color = gui.Color.from_text(color_str)
        if not color.isValid():
            raise TypeError()
        bitmap = gui.Pixmap(16, 16)
        bitmap.fill(color)
        return cls(bitmap)

    @classmethod
    def from_char(cls, char: str, background="black", color="white"):
        """Create a QIcon with a given character."""
        icon = cls()
        for size in (16, 32, 64):
            px = gui.Pixmap.create_char(
                char, background=background, color=color, size=size
            )
            icon.addPixmap(px)
        return icon

    @classmethod
    def from_image(cls, image: QtGui.QImage):
        return cls(gui.Pixmap.fromImage(image))

    def get_available_sizes(
        self, mode: ModeStr = "normal", state: StateStr = "off"
    ) -> list[core.Size]:
        if mode not in MODE:
            raise InvalidParamError(mode, MODE)
        if state not in STATE:
            raise InvalidParamError(state, STATE)
        return [core.Size(i) for i in self.availableSizes(MODE[mode], STATE[state])]

    def add_pixmap(
        self,
        data: QtCore.QByteArray | QtGui.QPixmap | bytes,
        mode: ModeStr = "normal",
        state: StateStr = "off",
    ):
        if mode not in MODE:
            raise InvalidParamError(mode, MODE)
        if state not in STATE:
            raise InvalidParamError(state, STATE)
        if isinstance(data, bytes):
            data = QtCore.QByteArray(data)
        if isinstance(data, QtCore.QByteArray):
            pixmap = QtGui.QPixmap()
            pixmap.loadFromData(data)
        else:
            pixmap = data
        self.addPixmap(pixmap, MODE[mode], STATE[state])

    def get_pixmap(
        self,
        size: types.SizeType | int,
        mode: ModeStr = "normal",
        state: StateStr = "off",
    ) -> QtGui.QPixmap:
        if mode not in MODE:
            raise InvalidParamError(mode, MODE)
        if state not in STATE:
            raise InvalidParamError(state, STATE)
        if isinstance(size, tuple):
            size = core.Size(*size)
        elif isinstance(size, int):
            size = core.Size(size, size)
        return self.pixmap(size, MODE[mode], STATE[state])

    def get_actual_size(
        self,
        size: types.SizeType | int,
        mode: ModeStr = "normal",
        state: StateStr = "off",
    ) -> core.Size:
        if mode not in MODE:
            raise InvalidParamError(mode, MODE)
        if state not in STATE:
            raise InvalidParamError(state, STATE)
        if isinstance(size, tuple):
            size = core.Size(*size)
        elif isinstance(size, int):
            size = core.Size(size, size)
        return core.Size(self.actualSize(size, MODE[mode], STATE[state]))


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    icon = Icon.from_char("A")
    window = widgets.MainWindow()
    window.set_icon(icon)
    window.show()
    app.main_loop()



from __future__ import annotations

from collections.abc import Mapping
from datetime import timedelta
import functools
import operator
import re
import sys
from typing import Any


REGEX = re.compile(
    r"((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?"
    r"((?P<seconds>\d+?)s)?((?P<milliseconds>\d+?)ms)?"
)


def dump_json(data: str):
    try:
        import orjson

        OPTS = orjson.OPT_NAIVE_UTC | orjson.OPT_SERIALIZE_NUMPY
        return orjson.dumps(data, option=OPTS)
    except (ImportError, ModuleNotFoundError):
        import json

        return json.dumps(data).encode()


def load_json(data):
    try:
        import orjson

        return orjson.loads(data)
    except (ImportError, ModuleNotFoundError):
        import json

        return json.loads(data)


def parse_time(time_str: str) -> int:
    parts = REGEX.match(time_str)
    if not parts:
        raise ValueError(time_str)
    dct = parts.groupdict()
    time_params = {name: int(param) for (name, param) in dct.items() if param}
    secs = timedelta(**time_params).total_seconds()
    return int(secs * 1000)


def format_seconds(seconds: float) -> str:
    m, s = divmod(int(seconds), 60)
    h, m = divmod(m, 60)
    return f"{h:02}:{h:02}:{h:02}"


def string_to_num_array(array: str) -> list[float]:
    floats = [float(i) for i in array.split(",")]
    return [int(i) if i.is_integer() else i for i in floats]


def merge_flags(flags, mapping: Mapping):
    return functools.reduce(operator.ior, [mapping[t] for t in flags])


def format_kwargs(kwargs: dict[str, Any]) -> str:
    kwarg_list = [f"{k}={repr(v)}" for k, v in kwargs.items()]
    return ", ".join(kwarg_list)


def cut_off_str(obj, max_len: int) -> str:
    """Create a string representation of an object, no longer than max_len characters.

    Uses repr(obj) to create the string representation. If this is longer than max_len -3
    characters, the last three will be replaced with elipsis.
    """
    s = repr(obj)
    if len(s) > max_len - 3:
        return s[: max_len - 3] + "..."
    return s


def get_color_percentage(
    color_1: tuple[int, int, int, int], color_2: tuple[int, int, int, int], percent: float
) -> tuple[int, int, int, int]:
    """Get a color which is percent% interpolated between start and end.

    Args:
        color_1 : Start color components (R, G, B, A / H, S, V, A / H, S, L, A)
        color_2 : End color components (R, G, B, A / H, S, V, A / H, S, L, A)
        percent: Percentage to interpolate, 0-100.
                 0: Start color will be returned.
                 100: End color will be returned.

    Return:
        A (x, y, z, alpha) tuple with the interpolated color components.
    """
    if not 0 <= percent <= 100:
        raise ValueError("percent needs to be between 0 and 100!")
    x = round(color_1[0] + (color_2[0] - color_1[0]) * percent / 100)
    y = round(color_1[1] + (color_2[1] - color_1[1]) * percent / 100)
    z = round(color_1[2] + (color_2[2] - color_1[2]) * percent / 100)
    a = round(color_1[3] + (color_2[3] - color_1[3]) * percent / 100)
    return (x, y, z, a)


def is_dark_mode() -> bool:
    if sys.platform.startswith("win"):
        from prettyqt import core

        path = (
            "HKEY_CURRENT_USER\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\"
            "Themes\\Personalize"
        )
        settings = core.Settings(path, core.Settings.Format.NativeFormat)
        return settings.value("AppsUseLightTheme") == 0
    elif sys.platform == "darwin":
        import darkdetect

        return darkdetect.isDark()
    else:
        return False



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import Callable, Literal

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict


BUTTONS = bidict(
    cancel=QtWidgets.QDialogButtonBox.StandardButton.Cancel,
    ok=QtWidgets.QDialogButtonBox.StandardButton.Ok,
    save=QtWidgets.QDialogButtonBox.StandardButton.Save,
    open=QtWidgets.QDialogButtonBox.StandardButton.Open,
    close=QtWidgets.QDialogButtonBox.StandardButton.Close,
    discard=QtWidgets.QDialogButtonBox.StandardButton.Discard,
    apply=QtWidgets.QDialogButtonBox.StandardButton.Apply,
    reset=QtWidgets.QDialogButtonBox.StandardButton.Reset,
    restore_defaults=QtWidgets.QDialogButtonBox.StandardButton.RestoreDefaults,
    help=QtWidgets.QDialogButtonBox.StandardButton.Help,
    save_all=QtWidgets.QDialogButtonBox.StandardButton.SaveAll,
    yes=QtWidgets.QDialogButtonBox.StandardButton.Yes,
    yes_to_all=QtWidgets.QDialogButtonBox.StandardButton.YesToAll,
    no=QtWidgets.QDialogButtonBox.StandardButton.No,
    no_to_all=QtWidgets.QDialogButtonBox.StandardButton.NoToAll,
    abort=QtWidgets.QDialogButtonBox.StandardButton.Abort,
    retry=QtWidgets.QDialogButtonBox.StandardButton.Retry,
    ignore=QtWidgets.QDialogButtonBox.StandardButton.Ignore,
)

ButtonStr = Literal[
    "cancel",
    "ok",
    "save",
    "open",
    "close",
    "discard",
    "apply",
    "reset",
    "restore_defaults",
    "help",
    "save_all",
    "yes",
    "yes_to_all",
    "no",
    "no_to_all",
    "abort",
    "retry",
    "ignore",
]

ROLES = bidict(
    invalid=QtWidgets.QDialogButtonBox.ButtonRole.InvalidRole,
    accept=QtWidgets.QDialogButtonBox.ButtonRole.AcceptRole,
    reject=QtWidgets.QDialogButtonBox.ButtonRole.RejectRole,
    destructive=QtWidgets.QDialogButtonBox.ButtonRole.DestructiveRole,
    action=QtWidgets.QDialogButtonBox.ButtonRole.ActionRole,
    help=QtWidgets.QDialogButtonBox.ButtonRole.HelpRole,
    yes=QtWidgets.QDialogButtonBox.ButtonRole.YesRole,
    no=QtWidgets.QDialogButtonBox.ButtonRole.NoRole,
    apply=QtWidgets.QDialogButtonBox.ButtonRole.ApplyRole,
    reset=QtWidgets.QDialogButtonBox.ButtonRole.ResetRole,
)

RoleStr = Literal[
    "invalid",
    "accept",
    "reject",
    "destructive",
    "action",
    "help",
    "yes",
    "no",
    "apply",
    "reset",
]

QtWidgets.QDialogButtonBox.__bases__ = (widgets.Widget,)


class DialogButtonBox(QtWidgets.QDialogButtonBox):

    button_clicked = core.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clicked.connect(self.on_click)

    def __len__(self) -> int:
        return len(self.buttons())

    def __getitem__(self, index: ButtonStr) -> QtWidgets.QPushButton:
        return self.button(BUTTONS[index])

    def __iter__(self) -> Iterator[QtWidgets.QAbstractButton]:
        return iter(self.buttons())

    def __contains__(self, index: ButtonStr):
        return self.button(BUTTONS[index]) is not None

    @classmethod
    def create(cls, **kwargs):
        box = cls()
        for k, v in kwargs.items():
            box.add_default_button(k, callback=v)  # type: ignore
        return box

    def on_click(self, button: QtCore.QObject):
        self.button_clicked.emit(button.objectName())

    def set_horizontal(self):
        self.setOrientation(constants.HORIZONTAL)

    def set_vertical(self):
        self.setOrientation(constants.VERTICAL)

    def set_orientation(self, orientation: constants.OrientationStr):
        """Set the orientation of the button box.

        Args:
            orientation: orientation for the button box

        Raises:
            InvalidParamError: orientation does not exist
        """
        if orientation not in constants.ORIENTATION:
            raise InvalidParamError(orientation, constants.ORIENTATION)
        self.setOrientation(constants.ORIENTATION[orientation])

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]

    def add_default_buttons(self, buttons: Sequence[ButtonStr]):
        return [self.add_default_button(btn) for btn in buttons]

    def add_default_button(
        self, button: ButtonStr, callback: Callable | None = None
    ) -> QtWidgets.QPushButton:
        """Add a default button.

        Args:
            button: button to add
            callback: function to call when button gets clicked

        Returns:
            created button

        Raises:
            InvalidParamError: Button type not available
        """
        if button not in BUTTONS:
            raise InvalidParamError(button, BUTTONS)
        btn = self.addButton(BUTTONS[button])
        btn.setObjectName(button)
        if callback:
            btn.clicked.connect(callback)
        return btn

    def add_button(
        self,
        button: QtWidgets.QPushButton | ButtonStr,
        role: RoleStr = "accept",
        callback: Callable | None = None,
    ) -> widgets.PushButton:
        """Add a button.

        Args:
            button: button to add
            role: role of the button
            callback: function to call when button gets clicked

        Returns:
            created button

        Raises:
            InvalidParamError: Button type not available
        """
        if isinstance(button, str):
            button = widgets.PushButton(button)
        self.addButton(button, ROLES[role])
        if callback:
            button.clicked.connect(callback)
        return button


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = DialogButtonBox()
    buttons = list(BUTTONS.keys())
    widget.add_default_buttons(buttons)
    widget.button_clicked.connect(print)
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import statemachine
from prettyqt.qt import QtStateMachine


QtStateMachine.QKeyEventTransition.__bases__ = (statemachine.EventTransition,)


class KeyEventTransition(QtStateMachine.QKeyEventTransition):
    pass



from typing import Optional

from prettyqt import core
from prettyqt.qt import QtCore


class PropertyMeta(type(QtCore.QObject)):  # type: ignore
    def __new__(cls, name, bases, attrs):
        for key in list(attrs.keys()):
            attr = attrs[key]
            if not isinstance(attr, SyncedProperty):
                continue
            initial_value = attr.initial_value
            type_ = type(initial_value)
            notifier = core.Signal(type_)
            attrs[key] = PropertyImpl(
                initial_value, name=key, type_=type_, notify=notifier
            )
            attrs[signal_attribute_name(key)] = notifier
        return super().__new__(cls, name, bases, attrs)


class SyncedProperty:
    """Property definition.

    This property will be patched by the PropertyMeta metaclass into a PropertyImpl type.
    """

    def __init__(self, initial_value, name: str = ""):
        self.initial_value = initial_value
        self.name = name


class PropertyImpl(core.Property):
    """Actual property implementation using a signal to notify any change."""

    def __init__(
        self, initial_value, name: str = "", type_: Optional[type] = None, notify=None
    ):
        super().__init__(type_, self.getter, self.setter, notify=notify)
        self.initial_value = initial_value
        self.name = name

    def getter(self, inst):
        return getattr(inst, value_attribute_name(self.name), self.initial_value)

    def setter(self, inst, value):
        setattr(inst, value_attribute_name(self.name), value)
        notifier_signal = getattr(inst, signal_attribute_name(self.name))
        notifier_signal.emit(value)


def signal_attribute_name(property_name: str) -> str:
    """Return a magic key for the attribute storing the signal name."""
    return f"_{property_name}_prop_signal_"


def value_attribute_name(property_name: str) -> str:
    """Return a magic key for the attribute storing the property value."""
    return f"_{property_name}_prop_value_"


if __name__ == "__main__":

    class Demo(core.Object, metaclass=PropertyMeta):
        my_prop = SyncedProperty(3.14)

    Demo()



from __future__ import annotations

from prettyqt import statemachine
from prettyqt.qt import QtStateMachine


QtStateMachine.QEventTransition.__bases__ = (statemachine.AbstractTransition,)


class EventTransition(QtStateMachine.QEventTransition):
    pass



from __future__ import annotations

from prettyqt import core, positioning
from prettyqt.qt import QtPositioning


class GeoLocation(QtPositioning.QGeoLocation):
    def get_address(self) -> positioning.GeoAddress:
        return positioning.GeoAddress(self.address())

    def get_coordinate(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.coordinate())

    def get_bounding_shape(self) -> positioning.GeoShape:
        if core.VersionNumber.get_qt_version() < (6, 0, 0):
            return positioning.GeoRectangle(self.boundingBox())
        else:
            return positioning.GeoShape(self.boundingShape())


if __name__ == "__main__":
    location = GeoLocation()



"""Multimediawidgets module."""

from .videowidgetcontrol import VideoWidgetControl
from .videowidget import VideoWidget
from .graphicsvideoitem import GraphicsVideoItem

__all__ = ["VideoWidgetControl", "VideoWidget", "GraphicsVideoItem"]



"""Provides QtWinExtras classes and functions."""

from prettyqt.qt import PYQT5, PythonQtError


if PYQT5:
    from PyQt5.QtWinExtras import *
else:
    raise PythonQtError("No Qt bindings could be found")



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtCore, QtNetwork
from prettyqt.utils import InvalidParamError, bidict, types


CONTENT_TYPES = bidict(
    mixed=QtNetwork.QHttpMultiPart.ContentType.MixedType,
    related=QtNetwork.QHttpMultiPart.ContentType.RelatedType,
    form=QtNetwork.QHttpMultiPart.ContentType.FormDataType,
    alternative=QtNetwork.QHttpMultiPart.ContentType.AlternativeType,
)

ContentTypeStr = Literal["mixed", "related", "form", "alternative"]

QtNetwork.QHttpMultiPart.__bases__ = (core.Object,)


class HttpMultiPart(QtNetwork.QHttpMultiPart):
    def __add__(self, other: QtNetwork.QHttpPart):
        self.append(other)
        return self

    def set_content_type(self, typ: ContentTypeStr):
        """Set content type.

        Args:
            typ: content type

        Raises:
            InvalidParamError: content type does not exist
        """
        if typ not in CONTENT_TYPES:
            raise InvalidParamError(typ, CONTENT_TYPES)
        self.setContentType(CONTENT_TYPES[typ])

    def set_boundary(self, boundary: types.ByteArrayType):
        if isinstance(boundary, str):
            boundary = boundary.encode()
        if isinstance(boundary, bytes):
            boundary = QtCore.QByteArray(boundary)
        self.setBoundary(boundary)

    def get_boundary(self) -> str:
        return bytes(self.boundary()).decode()


if __name__ == "__main__":
    part = HttpMultiPart()



from __future__ import annotations

from typing import Literal

from prettyqt import statemachine
from prettyqt.qt import QtStateMachine
from prettyqt.utils import InvalidParamError, bidict


HISTORY_TYPE = bidict(
    shallow=QtStateMachine.QHistoryState.HistoryType.ShallowHistory,
    deep=QtStateMachine.QHistoryState.HistoryType.DeepHistory,
)

HistoryTypeStr = Literal["shallow", "deep"]

QtStateMachine.QHistoryState.__bases__ = (statemachine.AbstractState,)


class HistoryState(QtStateMachine.QHistoryState):
    def set_history_type(self, typ: HistoryTypeStr):
        """Set history type to use.

        Args:
            typ: history type to use

        Raises:
            InvalidParamError: history type does not exist
        """
        if typ not in HISTORY_TYPE:
            raise InvalidParamError(typ, HISTORY_TYPE)
        self.setHistoryType(HISTORY_TYPE[typ])

    def get_history_type(self) -> HistoryTypeStr:
        """Return current history type.

        Returns:
            history type
        """
        return HISTORY_TYPE.inverse[self.historyType()]


if __name__ == "__main__":
    state = HistoryState()



from typing import Callable

from prettyqt import constants


class ModelMixin:

    DATA_ROLE = constants.USER_ROLE
    DTYPE_ROLE = constants.USER_ROLE + 1  # type: ignore
    NAME_ROLE = constants.USER_ROLE + 2  # type: ignore
    SORT_ROLE = constants.USER_ROLE + 3  # type: ignore
    MAX_ROWS = 1_000_000
    HEADER = ["Name"]
    DEFAULT_FLAGS = (
        constants.DRAG_ENABLED  # type: ignore
        | constants.IS_ENABLED
        | constants.IS_SELECTABLE
        | constants.NO_CHILDREN
    )
    LABELS: dict = dict()
    CHECKSTATE: dict = dict()
    TOOLTIPS: dict = dict()
    DECORATIONS: dict = dict()
    SET_DATA: dict = dict()
    content_type = ""
    data_by_index: Callable
    update_row: Callable

    def headerData(self, offset: int, orientation, role):
        if role == constants.DISPLAY_ROLE:
            if orientation == constants.HORIZONTAL:
                return self.HEADER[offset]

    def columnCount(self, parent=None):
        return len(self.HEADER)

    def flags(self, index):
        """Required override for AbstractitemModels.

        returns corresponding flags for cell of supplied index
        """
        if not index.isValid():
            return constants.DROP_ENABLED
        if index.column() in self.SET_DATA:
            return self.DEFAULT_FLAGS | constants.IS_EDITABLE
        return self.DEFAULT_FLAGS

    def data(self, index, role=constants.DISPLAY_ROLE):
        if not index.isValid():
            return None
        item = self.data_by_index(index)
        if role == constants.DECORATION_ROLE:
            fn = self.DECORATIONS.get(index.column())
            if fn:
                return fn(item)
        elif role in [constants.DISPLAY_ROLE, constants.EDIT_ROLE]:
            fn = self.LABELS.get(index.column())
            if fn:
                return fn(item)
        elif role == constants.TOOLTIP_ROLE:
            fn = self.TOOLTIPS.get(index.column())
            if fn:
                return fn(item)
        elif role == constants.CHECKSTATE_ROLE:
            fn = self.CHECKSTATE.get(index.column())
            if fn:
                return fn(item)
        elif role == self.DATA_ROLE:
            return item
        return None

    def setData(self, index, value, role):
        if role == constants.EDIT_ROLE:
            if not value:
                return False
            item = self.data_by_index(index)
            fn = self.SET_DATA.get(index.column())
            if fn:
                fn(item, value)
                self.update_row(index.row())
                return True



from __future__ import annotations

from prettyqt import constants, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError


QtWidgets.QSplitterHandle.__bases__ = (widgets.Widget,)


class SplitterHandle(QtWidgets.QSplitterHandle):
    def __init__(
        self,
        orientation: constants.OrientationStr | QtCore.Qt.Orientation,
        parent: QtWidgets.QSplitter,
    ):
        if isinstance(orientation, QtCore.Qt.Orientation):
            ori = orientation
        else:
            ori = constants.ORIENTATION[orientation]
        super().__init__(ori, parent)

    def set_orientation(self, orientation: constants.OrientationStr):
        """Set the orientation of the slider.

        Args:
            orientation: orientation for the slider

        Raises:
            InvalidParamError: orientation does not exist
        """
        if orientation not in constants.ORIENTATION:
            raise InvalidParamError(orientation, constants.ORIENTATION)
        self.setOrientation(constants.ORIENTATION[orientation])

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.Splitter()
    handle = SplitterHandle("horizontal", w)
    handle.show()
    app.main_loop()



from __future__ import annotations

import os
from typing import Literal

from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import types


FormatStr = Literal["plaintext", "HTML", "markdown", "ODF"]


class TextDocumentWriter(QtGui.QTextDocumentWriter):
    def __repr__(self):
        return f"{type(self).__name__}({self.device()!r}, {self.format()!r})"

    def get_format(self) -> FormatStr:
        return bytes(self.format()).decode()  # type: ignore

    def set_format(self, fmt: FormatStr | bytes | QtCore.QByteArray):
        new = fmt.encode() if isinstance(fmt, str) else fmt
        self.setFormat(new)

    def set_file_name(self, name: types.PathType):
        path = name if isinstance(name, str) else os.fspath(name)
        self.setFileName(path)

    @classmethod
    def get_supported_document_formats(cls) -> list[str]:
        return [bytes(i).decode() for i in cls.supportedDocumentFormats()]


if __name__ == "__main__":
    doc = TextDocumentWriter("a.text", b"plaintext")
    doc.set_format("UTF-8")
    print(doc.get_format())



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui
from prettyqt.qt import QtGui
from prettyqt.utils import InvalidParamError, bidict, colors, types


ROLE = bidict(
    window=QtGui.QPalette.ColorRole.Window,  # same as Background
    window_text=QtGui.QPalette.ColorRole.WindowText,  # same as Foreground
    base=QtGui.QPalette.ColorRole.Base,
    alternate_base=QtGui.QPalette.ColorRole.AlternateBase,
    tool_tip_base=QtGui.QPalette.ColorRole.ToolTipBase,
    tool_tip_text=QtGui.QPalette.ColorRole.ToolTipText,
    placeholder_text=QtGui.QPalette.ColorRole.PlaceholderText,
    text=QtGui.QPalette.ColorRole.Text,
    button=QtGui.QPalette.ColorRole.Button,
    button_text=QtGui.QPalette.ColorRole.ButtonText,
    bright_text=QtGui.QPalette.ColorRole.BrightText,
    light=QtGui.QPalette.ColorRole.Light,
    midlight=QtGui.QPalette.ColorRole.Midlight,
    dark=QtGui.QPalette.ColorRole.Dark,
    mid=QtGui.QPalette.ColorRole.Mid,
    shadow=QtGui.QPalette.ColorRole.Shadow,
    highlight=QtGui.QPalette.ColorRole.Highlight,
    highlighted_text=QtGui.QPalette.ColorRole.HighlightedText,
    link=QtGui.QPalette.ColorRole.Link,
    link_visited=QtGui.QPalette.ColorRole.LinkVisited,
    none=QtGui.QPalette.ColorRole.NoRole,
)

RoleStr = Literal[
    "window",
    "window_text",
    "base",
    "alternate_base",
    "tool_tip_base",
    "tool_tip_text",
    "placeholder_text",
    "text",
    "button",
    "button_text",
    "bright_text",
    "light",
    "midlight",
    "dark",
    "mid",
    "shadow",
    "highlight",
    "highlighted_text",
    "link",
    "link_visited",
    "none",
]

GROUP = bidict(
    disabled=QtGui.QPalette.ColorGroup.Disabled,
    active=QtGui.QPalette.ColorGroup.Active,  # normal
    inactive=QtGui.QPalette.ColorGroup.Inactive,
)

GroupStr = Literal["disabled", "active", "inactive"]


class Palette(QtGui.QPalette):
    def __getstate__(self):
        return bytes(self)

    def __setstate__(self, ba):
        core.DataStream.write_bytearray(ba, self)

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __getitem__(self, index: RoleStr) -> gui.Color:
        return self.get_color(index)

    def __setitem__(self, index: RoleStr, value: types.ColorType):
        self.set_color(index, value)

    def __bytes__(self):
        ba = core.DataStream.create_bytearray(self)
        return bytes(ba)

    def __repr__(self):
        return f"{type(self).__name__}({self['button']}, {self['window']})"

    def highlight_inactive(self):
        color = self.color(self.ColorGroup.Active, self.ColorRole.Highlight)
        self.setColor(self.ColorGroup.Inactive, self.ColorRole.Highlight, color)

    def set_color(
        self, role: RoleStr, color: types.ColorType, group: GroupStr = "active"
    ):
        color = colors.get_color(color)
        self.setColor(GROUP[group], ROLE[role], color)

    def get_colors(self, group: GroupStr = "active") -> dict[str, gui.Color]:
        return {k: self.get_color(k, group) for k in ROLE.keys()}

    def get_color(self, role: RoleStr, group: GroupStr = "active") -> gui.Color:
        return gui.Color(self.color(GROUP[group], ROLE[role]))

    def set_brush(
        self, role: RoleStr, brush: types.ColorAndBrushType, group: GroupStr = "active"
    ):
        if not isinstance(brush, QtGui.QBrush):
            brush = gui.Brush(colors.get_color(brush))
        self.setBrush(GROUP[group], ROLE[role], brush)

    def get_brushes(self, group: GroupStr = "active") -> dict[str, gui.Brush]:
        return {k: self.get_brush(k, group) for k in ROLE.keys()}

    def get_brush(self, role: RoleStr, group: GroupStr = "active") -> gui.Brush:
        return gui.Brush(self.brush(GROUP[group], ROLE[role]))

    def set_color_group(self, group: GroupStr, *args, **kwargs):
        """Set the color group.

        Args:
            group: color group to use

        Raises:
            InvalidParamError: invalid color group
        """
        if group not in GROUP:
            raise InvalidParamError(group, GROUP)
        self.setColorGroup(GROUP[group], *args, **kwargs)

    def get_color_group(self) -> GroupStr:
        """Return color group.

        Returns:
            color group
        """
        return GROUP.inverse[self.colorGroup()]

    def inverted(self) -> Palette:
        pal = Palette()
        for group in GROUP:
            for role in ROLE:
                color = self.get_color(role, group)
                pal.set_color(role, color.inverted(), group)
        return pal

    @classmethod
    def create_dark_palette(cls) -> Palette:
        pal = cls()
        pal.set_color("window", gui.Color(53, 53, 53))
        pal.set_color("window_text", "white")
        pal.set_color("window_text", "grey", group="disabled")
        pal.set_color("base", gui.Color(25, 25, 25))
        pal.set_color("alternate_base", gui.Color(53, 53, 53))
        pal.set_color("tool_tip_base", "slategrey")
        pal.set_color("tool_tip_base", "slategrey", group="inactive")
        pal.set_color("tool_tip_text", "white")
        pal.set_color("tool_tip_text", "white", group="inactive")
        pal.set_color("text", "white")
        pal.set_color("text", "grey", group="disabled")
        pal.set_color("button", gui.Color(53, 53, 53))
        pal.set_color("button_text", "white")
        pal.set_color("button_text", "grey", group="disabled")
        pal.set_color("bright_text", "red")
        pal.set_color("link", "dodgerblue")
        pal.set_color("highlight", "dodgerblue")
        pal.set_color("highlight", gui.Color(80, 80, 80), group="disabled")
        pal.set_color("highlighted_text", "black")
        pal.set_color("highlighted_text", "grey", group="disabled")
        return pal


if __name__ == "__main__":
    pal = Palette()
    print(pal.get_colors())



from __future__ import annotations

from prettyqt.qt import QtCore


class MetaProperty:
    def __init__(self, metaproperty: QtCore.QMetaProperty):
        self.item = metaproperty

    def __bool__(self):
        return self.item.isValid()

    def __repr__(self):
        return f"{type(self).__name__}({self.get_name()!r})"

    def get_name(self) -> str:
        return self.item.name()  # type: ignore


if __name__ == "__main__":
    from prettyqt import core

    metaobj = core.Object.get_metaobject()



"""svg module.

contains QtSvg-based classes
"""

from prettyqt import core

from .svggenerator import SvgGenerator
from .svgrenderer import SvgRenderer

if core.VersionNumber.get_qt_version() < (6, 0, 0):
    from .graphicssvgitem import GraphicsSvgItem
    from .svgwidget import SvgWidget


__all__ = ["GraphicsSvgItem", "SvgGenerator", "SvgRenderer", "SvgWidget"]



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui, quick
from prettyqt.qt import QtQuick
from prettyqt.utils import InvalidParamError, bidict


mod = QtQuick.QQuickPaintedItem

PERFORMANCE_HINT = bidict(
    fast_fbo_resizing=mod.PerformanceHint.FastFBOResizing,
)

PerformanceHintStr = Literal["fast_fbo_resizing"]

RENDER_TARGET = bidict(
    image=mod.RenderTarget.Image,
    framebuffer_object=mod.RenderTarget.FramebufferObject,
    inverted_y_framebuffer_object=mod.RenderTarget.InvertedYFramebufferObject,
)

RenderTargetStr = Literal["image", "framebuffer_object" "inverted_y_framebuffer_object"]

QtQuick.QQuickPaintedItem.__bases__ = (quick.QuickItem,)


class QuickPaintedItem(QtQuick.QQuickPaintedItem):
    def get_fill_color(self) -> gui.Color:
        return gui.Color(self.fillColor())

    def get_texture_size(self) -> core.Size:
        return core.Size(self.textureSize())

    def set_render_target(self, target: RenderTargetStr):
        """Set the render target.

        Args:
            target: render target to use

        Raises:
            InvalidParamError: render target does not exist
        """
        if target not in RENDER_TARGET:
            raise InvalidParamError(target, RENDER_TARGET)
        self.setRenderTarget(RENDER_TARGET[target])

    def get_render_target(self) -> RenderTargetStr:
        """Return the render target.

        Returns:
            render target
        """
        return RENDER_TARGET.inverse[self.renderTarget()]



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import bidict


BUTTON_FEATURES = bidict(
    none=QtWidgets.QStyleOptionButton.ButtonFeature.None_,
    flat=QtWidgets.QStyleOptionButton.ButtonFeature.Flat,
    has_menu=QtWidgets.QStyleOptionButton.ButtonFeature.HasMenu,
    default_button=QtWidgets.QStyleOptionButton.ButtonFeature.DefaultButton,
    auto_default_button=QtWidgets.QStyleOptionButton.ButtonFeature.AutoDefaultButton,
    command_link_button=QtWidgets.QStyleOptionButton.ButtonFeature.CommandLinkButton,
)


QtWidgets.QStyleOptionButton.__bases__ = (widgets.StyleOption,)


class StyleOptionButton(QtWidgets.QStyleOptionButton):
    pass



from __future__ import annotations

import datetime

from prettyqt import core, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import types

QtWidgets.QDateEdit.__bases__ = (widgets.DateTimeEdit,)


class DateEdit(QtWidgets.QDateEdit):

    value_changed = core.Signal(datetime.datetime)

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setDate(state["date"])
        self.set_range(*state["range"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def serialize_fields(self):
        return dict(
            date=self.get_date(),
            range=(self.min_date(), self.max_date()),
        )

    def set_value(self, value: types.DateType):
        if isinstance(value, str):
            value = QtCore.QDate.fromString(value)
        self.setDate(value)  # type: ignore

    def set_range(self, lower: types.DateType, upper: types.DateType):
        if isinstance(lower, str):
            lower = QtCore.QDate.fromString(lower)
        if isinstance(upper, str):
            upper = QtCore.QDate.fromString(upper)
        self.setToolTip(f"{lower} <= x <= {upper}")
        self.setDateRange(lower, upper)  # type: ignore

    def get_value(self) -> datetime.date:
        return self.get_date()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = DateEdit()
    widget.show()
    app.main_loop()



# from prettyqt.qt import QtWebEngineCore

from __future__ import annotations

from typing import Literal

from prettyqt import core, webenginecore
from prettyqt.qt import QtWebEngineCore
from prettyqt.utils import InvalidParamError, bidict


mod = QtWebEngineCore.QWebEngineProfile

HTTP_CACHE_TYPE = bidict(
    none=mod.HttpCacheType.NoCache,
    disk=mod.HttpCacheType.DiskHttpCache,
    memory=mod.HttpCacheType.MemoryHttpCache,
)

HttpCacheTypeStr = Literal["none", "disk", "memory"]

PERSISTENT_COOKIE_POLICY = bidict(
    none=mod.PersistentCookiesPolicy.NoPersistentCookies,
    allow=mod.PersistentCookiesPolicy.AllowPersistentCookies,
    force=mod.PersistentCookiesPolicy.ForcePersistentCookies,
)

PersistentCookiePolicyStr = Literal["none", "allow", "force"]

QtWebEngineCore.QWebEngineProfile.__bases__ = (core.Object,)


class WebEngineProfile(QtWebEngineCore.QWebEngineProfile):
    def set_persistent_cookie_policy(self, policy: PersistentCookiePolicyStr):
        """Set the persistent cookie policy.

        Args:
            policy: persistent cookie policy

        Raises:
            InvalidParamError: Policy does not exist
        """
        if policy not in PERSISTENT_COOKIE_POLICY:
            raise InvalidParamError(policy, PERSISTENT_COOKIE_POLICY)
        self.setPersistentCookiesPolicy(PERSISTENT_COOKIE_POLICY[policy])

    def get_persistent_cookie_policy(self) -> PersistentCookiePolicyStr:
        """Return current persistent cookie policy.

        Returns:
            Persistent cookie policy
        """
        return PERSISTENT_COOKIE_POLICY.inverse[self.persistentCookiesPolicy()]

    def set_http_cache_type(self, typ: HttpCacheTypeStr):
        """Set the http cache type.

        Args:
            typ: http cache type

        Raises:
            InvalidParamError: Cache type does not exist
        """
        if typ not in HTTP_CACHE_TYPE:
            raise InvalidParamError(typ, HTTP_CACHE_TYPE)
        self.setHttpCacheType(HTTP_CACHE_TYPE[typ])

    def get_http_cache_type(self) -> HttpCacheTypeStr:
        """Return current http cache type.

        Returns:
            Http cache type
        """
        return HTTP_CACHE_TYPE.inverse[self.httpCacheType()]

    def get_scripts(self) -> webenginecore.WebEngineScriptCollection:
        return webenginecore.WebEngineScriptCollection(self.scripts())


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    item = WebEngineProfile()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtCore, QtHelp
from prettyqt.utils import types


QtHelp.QHelpContentWidget.__bases__ = (widgets.TreeView,)


class HelpContentWidget(QtHelp.QHelpContentWidget):
    def index_of(self, url: types.UrlType) -> QtCore.QModelIndex | None:
        if isinstance(url, str):
            url = QtCore.QUrl(url)
        idx = self.indexOf(url)
        if not idx.isValid():
            return None
        return idx



from __future__ import annotations

import sys
import traceback
from typing import Literal

from prettyqt import gui, iconprovider, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict, types


ICONS = bidict(
    none=QtWidgets.QMessageBox.Icon.NoIcon,
    information=QtWidgets.QMessageBox.Icon.Information,
    warning=QtWidgets.QMessageBox.Icon.Warning,
    critical=QtWidgets.QMessageBox.Icon.Critical,
    question=QtWidgets.QMessageBox.Icon.Question,
)

IconStr = Literal["none", "information", "warning", "critical", "question"]

BUTTONS = bidict(
    none=QtWidgets.QMessageBox.StandardButton.NoButton,
    cancel=QtWidgets.QMessageBox.StandardButton.Cancel,
    ok=QtWidgets.QMessageBox.StandardButton.Ok,
    save=QtWidgets.QMessageBox.StandardButton.Save,
    open=QtWidgets.QMessageBox.StandardButton.Open,
    close=QtWidgets.QMessageBox.StandardButton.Close,
    discard=QtWidgets.QMessageBox.StandardButton.Discard,
    apply=QtWidgets.QMessageBox.StandardButton.Apply,
    reset=QtWidgets.QMessageBox.StandardButton.Reset,
    restore_defaults=QtWidgets.QMessageBox.StandardButton.RestoreDefaults,
    help=QtWidgets.QMessageBox.StandardButton.Help,
    save_all=QtWidgets.QMessageBox.StandardButton.SaveAll,
    yes=QtWidgets.QMessageBox.StandardButton.Yes,
    yes_to_all=QtWidgets.QMessageBox.StandardButton.YesToAll,
    no=QtWidgets.QMessageBox.StandardButton.No,
    no_to_all=QtWidgets.QMessageBox.StandardButton.NoToAll,
    abort=QtWidgets.QMessageBox.StandardButton.Abort,
    retry=QtWidgets.QMessageBox.StandardButton.Retry,
    ignore=QtWidgets.QMessageBox.StandardButton.Ignore,
)

ButtonStr = Literal[
    "none",
    "cancel",
    "ok",
    "save",
    "open",
    "close",
    "discard",
    "apply",
    "reset",
    "restore_defaults",
    "help",
    "save_all",
    "yes",
    "yes_to_all",
    "no",
    "no_to_all",
    "abort",
    "retry",
    "ignore",
]

BUTTON_ROLE = bidict(
    invalid=QtWidgets.QMessageBox.ButtonRole.InvalidRole,
    accept=QtWidgets.QMessageBox.ButtonRole.AcceptRole,
    reject=QtWidgets.QMessageBox.ButtonRole.RejectRole,
    destructive=QtWidgets.QMessageBox.ButtonRole.DestructiveRole,
    action=QtWidgets.QMessageBox.ButtonRole.ActionRole,
    help=QtWidgets.QMessageBox.ButtonRole.HelpRole,
    yes=QtWidgets.QMessageBox.ButtonRole.YesRole,
    no=QtWidgets.QMessageBox.ButtonRole.NoRole,
    apply=QtWidgets.QMessageBox.ButtonRole.ApplyRole,
    reset=QtWidgets.QMessageBox.ButtonRole.ResetRole,
)

ButtonRoleStr = Literal[
    "invalid",
    "accept",
    "reject",
    "destructive",
    "action",
    "help",
    "yes",
    "no",
    "apply",
    "reset",
]

TEXT_FORMAT = bidict(
    rich=QtCore.Qt.TextFormat.RichText,
    plain=QtCore.Qt.TextFormat.PlainText,
    auto=QtCore.Qt.TextFormat.AutoText,
    markdown=QtCore.Qt.TextFormat.MarkdownText,
)

TextFormatStr = Literal["rich", "plain", "auto", "markdown"]

QtWidgets.QMessageBox.__bases__ = (widgets.Dialog,)


class MessageBox(QtWidgets.QMessageBox):
    def __init__(
        self,
        icon: types.IconType | IconStr = None,
        title: str = "",
        text: str = "",
        informative_text: str = "",
        details: str = "",
        buttons: list[ButtonStr] | None = None,
        parent: QtWidgets.QWidget | None = None,
    ):
        super().__init__(parent)
        self.set_icon(icon)
        self.setText(text)
        self.setInformativeText(informative_text)
        self.setWindowTitle(title)
        self.setWindowFlags(
            QtCore.Qt.WindowType.Dialog  # type: ignore
            | QtCore.Qt.WindowType.WindowTitleHint
            | QtCore.Qt.WindowType.CustomizeWindowHint
        )
        self.setDetailedText(details)
        if isinstance(buttons, list):
            for b in buttons:
                self.add_button(b)

    def serialize_fields(self):
        return dict(
            icon=self.get_icon(),
            detailed_text=self.detailedText(),
            icon_pixmap=self.get_icon_pixmap(),
            informative_text=self.informativeText(),
            text=self.text(),
            text_format=self.get_text_format(),
        )

    @classmethod
    def message(
        cls,
        text: str,
        title: str = "",
        icon: types.IconType = None,
        detail_text: str | None = None,
    ) -> str:
        m = cls("none", title, text)
        m.set_icon(icon)
        if detail_text is not None:
            m.setDetailedText(detail_text)
        return m.show_blocking()

    @classmethod
    def show_exception(cls, exception: Exception):
        exctype, value = sys.exc_info()[:2]
        tb = traceback.format_exc()
        dlg = cls(text=str(value), title=str(exctype), icon="critical", details=tb)
        dlg.show_blocking()

    def set_icon(self, icon: types.IconType | IconStr):
        if icon in ICONS:
            self.setIcon(ICONS[icon])
        else:
            ico = iconprovider.get_icon(icon)
            self.setIconPixmap(ico.get_pixmap(size=64))

    def show_blocking(self) -> ButtonStr:
        return BUTTONS.inverse[self.main_loop()]

    def get_icon_pixmap(self) -> gui.Pixmap | None:
        pix = self.iconPixmap()
        if pix.isNull():
            return None
        return gui.Pixmap(pix)

    def get_standard_buttons(self) -> list[ButtonStr]:
        return [k for k, v in BUTTONS.items() if v & self.standardButtons()]

    def add_button(self, button: ButtonStr) -> QtWidgets.QPushButton:
        """Add a default button.

        Args:
            button: button to add

        Returns:
            created button

        Raises:
            InvalidParamError: Button type not available
        """
        if button not in BUTTONS:
            raise InvalidParamError(button, BUTTONS)
        return self.addButton(BUTTONS[button])

    # @classmethod
    # def show_exception(cls, exception):
    #     header = str(exception[0])
    #     error_text = str(exception[1])
    #     widgets.MessageBox.message(error_text, header, "mdi.exclamation")

    def set_text_format(self, text_format: TextFormatStr):
        """Set the text format.

        Args:
            text_format: text format to use

        Raises:
            InvalidParamError: text format does not exist
        """
        if text_format not in TEXT_FORMAT:
            raise InvalidParamError(text_format, TEXT_FORMAT)
        self.setTextFormat(TEXT_FORMAT[text_format])

    def get_text_format(self) -> TextFormatStr:
        """Return current text format.

        Returns:
            text format
        """
        return TEXT_FORMAT.inverse[self.textFormat()]

    def set_escape_button(self, button: ButtonStr | QtWidgets.QAbstractButton):
        if isinstance(button, QtWidgets.QAbstractButton):
            self.setEscapeButton(button)
        else:
            self.setEscapeButton(BUTTONS[button])

    def set_default_button(self, button: ButtonStr | QtWidgets.QPushButton):
        if isinstance(button, QtWidgets.QPushButton):
            self.setDefaultButton(button)
        else:
            self.setDefaultButton(BUTTONS[button])


if __name__ == "__main__":
    app = widgets.app()
    ret = MessageBox(icon="warning", title="header", text="text", details="details")
    ret.set_icon("mdi.folder")
    ret.show()
    print(ret)
    app.main_loop()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsTransform.__bases__ = (core.Object,)


class GraphicsTransform(QtWidgets.QGraphicsTransform):
    pass


if __name__ == "__main__":
    transform = GraphicsTransform()



from __future__ import annotations

import contextlib

from prettyqt.qt import QtCore


class Semaphore(QtCore.QSemaphore):
    @contextlib.contextmanager
    def acquire_resources(self, n: int = 1):
        result = self.tryAcquire(n)
        yield result
        self.release(n)



from __future__ import annotations

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtWidgets


QtWidgets.QScrollBar.__bases__ = (widgets.AbstractSlider,)


class ScrollBar(QtWidgets.QScrollBar):

    value_changed = core.Signal(int)

    def __init__(
        self,
        orientation: (QtCore.Qt.Orientation | constants.OrientationStr) = "horizontal",
        parent: QtWidgets.QWidget | None = None,
    ):
        if isinstance(orientation, QtCore.Qt.Orientation):
            ori = orientation
        else:
            ori = constants.ORIENTATION[orientation]
        super().__init__(ori, parent)
        self.valueChanged.connect(self.on_value_change)



from __future__ import annotations

from prettyqt import gui, widgets
from prettyqt.qt import QtGui, QtWidgets
from prettyqt.utils import InvalidParamError


QtWidgets.QWizardPage.__bases__ = (widgets.Widget,)


class WizardPage(QtWidgets.QWizardPage):
    def serialize_fields(self):
        return dict(title=self.title(), sub_title=self.subTitle())

    def set_pixmap(
        self, typ: widgets.wizard.WizardPixmapStr, pixmap: QtGui.QPixmap | None
    ):
        if typ not in widgets.wizard.WIZARD_PIXMAP:
            raise InvalidParamError(typ, widgets.wizard.WIZARD_PIXMAP)
        if pixmap is None:
            pixmap = QtGui.QPixmap()
        self.setPixmap(widgets.wizard.WIZARD_PIXMAP[typ], pixmap)

    def get_pixmap(self, typ: widgets.wizard.WizardPixmapStr) -> gui.Pixmap | None:
        if typ not in widgets.wizard.WIZARD_PIXMAP:
            raise InvalidParamError(typ, widgets.wizard.WIZARD_PIXMAP)
        pix = gui.Pixmap(self.pixmap(widgets.wizard.WIZARD_PIXMAP[typ]))
        if pix.isNull():
            return None
        return pix

    def set_button_text(self, button_type: widgets.wizard.WizardButtonStr, value: str):
        """Set text for given button type.

        Args:
            button_type: button to get text from
            value: text to set

        """
        if button_type not in widgets.wizard.WIZARD_BUTTON:
            raise InvalidParamError(button_type, widgets.wizard.WIZARD_BUTTON)
        self.setButtonText(widgets.wizard.WIZARD_BUTTON[button_type], value)

    def get_button_text(self, button_type: widgets.wizard.WizardButtonStr) -> str:
        """Return text for given button type.

        Args:
            button_type: button to get text from

        Returns:
            Button text
        """
        if button_type not in widgets.wizard.WIZARD_BUTTON:
            raise InvalidParamError(button_type, widgets.wizard.WIZARD_BUTTON)
        return self.buttonText(widgets.wizard.WIZARD_BUTTON[button_type])
